@@ -14,9 +14,10 @@ import heapq
 import os.path
 from sys import platform
 
+from .consts import MAX_LABEL_COST
+
 
 __all__ = [
-    'MAX_LABEL_COST',
     'single_source_shortest_path',
     'output_path_sequence',
     'find_shortest_path',
@@ -24,10 +25,6 @@ __all__ = [
 ]
 
 
-# for initialization in shortest path calculation
-MAX_LABEL_COST = 10000
-
-
 if platform.startswith('win32'):
     _dll_file = os.path.join(os.path.dirname(__file__), 'bin/path_engine.dll')
 elif platform.startswith('linux'):
@@ -49,38 +46,42 @@ _cdll.shortest_path.argtypes = [
     ctypes.POINTER(ctypes.c_int),
     ctypes.POINTER(ctypes.c_int),
     ctypes.POINTER(ctypes.c_int),
+    ctypes.POINTER(ctypes.c_wchar_p),
     ctypes.POINTER(ctypes.c_double),
     ctypes.POINTER(ctypes.c_double),
     ctypes.POINTER(ctypes.c_int),
     ctypes.POINTER(ctypes.c_int),
     ctypes.POINTER(ctypes.c_int),
+    ctypes.c_char,
     ctypes.c_int,
     ctypes.c_int
 ]
 
 
-def _optimal_label_correcting_CAPI(G, origin_node_no,
-                                   departure_time=0,
-                                   first_thru_node=0):
+def _optimal_label_correcting_CAPI(G,
+                                   origin_node_no,
+                                   departure_time=0):
     """ call the deque implementation of MLC written in cpp
 
     node_label_cost, node_predecessor, and link_predecessor are still
     initialized in shortest_path() even the source node has no outgoing links.
     """
-    _cdll.shortest_path(origin_node_no,
+    _cdll.shortest_path_n(origin_node_no,
                         G.get_node_size(),
                         G.get_from_node_no_arr(),
                         G.get_to_node_no_arr(),
                         G.get_first_links(),
                         G.get_last_links(),
                         G.get_sorted_link_no_arr(),
+                        G.get_allowed_uses(),
                         G.get_link_costs(),
                         G.get_node_label_costs(),
                         G.get_node_preds(),
                         G.get_link_preds(),
                         G.get_queue_next(),
-                        departure_time,
-                        first_thru_node)
+                        G.get_agent_type_str(),
+                        G.get_last_thru_node(),
+                        departure_time)
 
 
 def _single_source_shortest_path_fifo(G, origin_node_no):
@@ -204,8 +205,8 @@ def _single_source_shortest_path_dijkstra(G, origin_node_no):
                 heapq.heappush(SEList, (G.node_label_cost[to_node], to_node))
 
 
-def single_source_shortest_path(G, origin_node_id, engine_type='c',
-                                sp_algm='deque'):
+def single_source_shortest_path(G, origin_node_id,
+                                engine_type='c', sp_algm='deque'):
 
     origin_node_no = G.get_node_no(origin_node_id)
 
@@ -239,7 +240,7 @@ def single_source_shortest_path(G, origin_node_id, engine_type='c',
                             +'fifo or deque or dijkstra')
 
 
-def output_path_sequence(G, from_node_id, to_node_id, type='node'):
+def output_path_sequence(G, to_node_id, type='node'):
     """ output shortest path in terms of node sequence or link sequence
 
     Note that this function returns GENERATOR rather than list.
@@ -260,8 +261,8 @@ def output_path_sequence(G, from_node_id, to_node_id, type='node'):
         current_link_seq_no = G.link_predecessor[current_node_seq_no]
         while current_link_seq_no >= 0:
             path.append(current_link_seq_no)
-            current_link_seq_no = G.link_predecessor[current_node_seq_no]
             current_node_seq_no = G.node_predecessor[current_node_seq_no]
+            current_link_seq_no = G.link_predecessor[current_node_seq_no]
         # reverse the sequence
         for link_seq_no in reversed(path):
             yield G.link_list[link_seq_no].get_link_id()
@@ -282,14 +283,15 @@ def find_shortest_path(G, from_node_id, to_node_id, seq_type='node'):
     single_source_shortest_path(G, from_node_id, engine_type='c')
 
     # return list(output_path_sequence(G, from_node_id, to_node_id, seq_type))
-    return ';'.join(
-        str(x) for x in output_path_sequence(G, from_node_id,
-                                             to_node_id, seq_type)
+    path = ';'.join(
+        str(x) for x in output_path_sequence(G, to_node_id, seq_type)
     )
 
+    return f'distance: {_get_path_cost(G, to_node_id):.2f} | path: {path}'
+
 
 def find_path_for_agents(G, column_pool, engine_type='c'):
-    """ find shortest path for each agent
+    """ find and set up shortest path for each agent
 
     the internal node and links will be used to set up the node sequence and
     link sequence respectively
